/*
 * imgRAS.c --
 *
 *	A photo image file handler for SUN Raster files.
 *
 * Written by:
 *	Jan Nijtmans
 *	NICI (Nijmegen Institute of Cognition and Information)
 *	email: nijtmans@nici.kun.nl
 *	url:   http://www.cogsci.kun.nl/~nijtmans/
 *
 */

#include "imgInt.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

/* constants used only in this file */

#define MAX_BUFFER 4096

/*
 * The format record for the RAS file format:
 */

static int		FileMatchRAS _ANSI_ARGS_((FILE *f, char *fileName,
			    char *formatString, int *widthPtr,
			    int *heightPtr));
static int      	StringMatchRAS _ANSI_ARGS_(( char *string,
		            char *formatString, int *widthPtr, int *heightPtr));
static int		FileReadRAS  _ANSI_ARGS_((Tcl_Interp *interp,
			    FILE *f, char *fileName, char *formatString,
			    Tk_PhotoHandle imageHandle, int destX, int destY,
			    int width, int height, int srcX, int srcY));
static int	        StringReadRAS _ANSI_ARGS_((Tcl_Interp *interp,
			    char *string, char *formatString,
              		    Tk_PhotoHandle imageHandle, int destX, int destY,
		            int width, int height, int srcX, int srcY));
static int	        StringWriteRAS _ANSI_ARGS_((Tcl_Interp *interp,
               		    Tcl_DString *dataPtr, char *formatString,
		            Tk_PhotoImageBlock *blockPtr));
static int              FileWriteRAS _ANSI_ARGS_((Tcl_Interp *interp,
                            char *fileName, char *formatString,
                            Tk_PhotoImageBlock *blockPtr));

Tk_PhotoImageFormat imgFmtRAS = {
    "RAS",			/* name */
    FileMatchRAS,		/* fileMatchProc */
    StringMatchRAS,		/* stringMatchProc */
    FileReadRAS,		/* fileReadProc */
    StringReadRAS,		/* stringReadProc */
    FileWriteRAS,		/* fileWriteProc */
    StringWriteRAS,		/* stringWriteProc */
};

/*
 * RAS's are represented as data in base64 format.
 * base64 strings consist of 4 6-bit characters -> 3 8 bit bytes.
 * A-Z, a-z, 0-9, + and / represent the 64 values (in order).
 * '=' is a trailing padding char when the un-encoded data is not a
 * multiple of 3 bytes.  We'll ignore white space when encountered.
 * Any other invalid character is treated as an EOF
 */

#define RAS_SPECIAL	(256)
#define RAS_PAD		(RAS_SPECIAL+1)
#define RAS_SPACE	(RAS_SPECIAL+2)
#define RAS_BAD		(RAS_SPECIAL+3)
#define RAS_DONE	(RAS_SPECIAL+4)

static char base64_table[64] = {
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
    'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
    'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
    'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
    'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
    'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
    'w', 'x', 'y', 'z', '0', '1', '2', '3',
    '4', '5', '6', '7', '8', '9', '+', '/'
};

static int fromData = 0;

/*
 * structure to "mimic" FILE for Mread, so we can look like fread.
 * The decoder state keeps track of which byte we are about to read,
 * or EOF.
 */

typedef struct mFile {
    Tcl_DString *buffer;	/* pointer to dynamical string for writing */
    char *data;			/* mmencoded source/destination string */
    int c;			/* bits left over from previous character */
    int state;			/* decoder state (0-4 or RAS_DONE) */
    int linelength;		/* length of phisical line already written */
} MFile;

/*
 * Prototypes for local procedures defined in this file:
 */

static int	ReadRASFileHeader _ANSI_ARGS_((FILE *f, 
			int *widthPtr, int *heightPtr, int *depthPtr,
			int *lengthPtr, int *typePtr,
			int *mapTypePtr, int *mapLengthPtr));
static int	ReadRAS _ANSI_ARGS_((Tcl_Interp *interp, FILE *f,
			char *fileName,
			char *formatString, Tk_PhotoHandle imageHandle,
			int destX, int destY, int width, int height,
			int srcX, int srcY));
static int	WriteRAS _ANSI_ARGS_((Tcl_Interp *interp,
			char *fileName, Tcl_DString *dataPtr,
			char *formatString,
			Tk_PhotoImageBlock *blockPtr));
/*
 * these are for the BASE64 image reader code only
 */

static int	Fread _ANSI_ARGS_((unsigned char *dst, size_t size,
		    size_t count, FILE *file));
static int	Mread _ANSI_ARGS_((char *dst, size_t size,
		    size_t count, MFile *handle));
static int	Mgetc _ANSI_ARGS_((MFile *handle));
static int	char64 _ANSI_ARGS_((int c));
static void	mInit _ANSI_ARGS_((char *string,
		    MFile *handle));
static int	Fwrite _ANSI_ARGS_((unsigned char *src, size_t size,
		    size_t count, FILE *file));
static int	Mwrite _ANSI_ARGS_((unsigned char *src, size_t size,
		    size_t count, MFile *handle));
static void	mWriteInit _ANSI_ARGS_((Tcl_DString *buffer,
		    MFile *handle));
static int	Fputc _ANSI_ARGS_((int c, FILE *file));
static int	Mputc _ANSI_ARGS_((int c, MFile *handle));

/*
 *----------------------------------------------------------------------
 *
 * StringMatchRAS --
 *
 *	This procedure is invoked by the photo image type to see if
 *	a datastring contains image data in RAS format.
 *
 * Results:
 *	The return value is >0 if the first characters in data look
 *	like RAS data, and 0 otherwise.
 *
 * Side effects:
 *	none
 *
 *----------------------------------------------------------------------
 */
static int
StringMatchRAS(string, formatString, widthPtr, heightPtr)
    char *string;               /* The data supplied by the image */
    char *formatString;		/* User-specified format string, or NULL. */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here if the file is a valid
				 * raw RAS file. */
{
    MFile handle;
    int dept, length, type, mapType, mapLength, result;

    mInit(string,&handle);
    fromData = 1;
    result = ReadRASFileHeader((FILE *) &handle, widthPtr, heightPtr,
	&dept, &length, &type, &mapType, &mapLength);
    fromData = 0;
    return result;
}


/*
 *----------------------------------------------------------------------
 *
 * FileMatchRAS --
 *
 *	This procedure is invoked by the photo image type to see if
 *	a file contains image data in RAS format.
 *
 * Results:
 *	The return value is >0 if the first characters in file "f" look
 *	like RAS data, and 0 otherwise.
 *
 * Side effects:
 *	The access position in f may change.
 *
 *----------------------------------------------------------------------
 */

static int
FileMatchRAS(f, fileName, formatString, widthPtr, heightPtr)
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here if the file is a valid
				 * raw RAS file. */
{
    int dept, length, type, mapType, mapLength;

    return ReadRASFileHeader(f, widthPtr, heightPtr, &dept, &length, &type,
	&mapType, &mapLength);
}


/*
 *----------------------------------------------------------------------
 *
 * ReadRAS --
 *
 *	This procedure is called by the photo image type to read
 *	RAS format data from a file or string and write it into a
 *	given photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	The access position in file f is changed (if read from file)
 *	and new data is added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

typedef struct myblock {
    Tk_PhotoImageBlock pub;
    int dummy; /* extra RAS_SPACE for offset[3], if not included already
		  in Tk_PhotoImageBlock */
} myblock;

static int
ReadRAS(interp, f, fileName, dataPtr, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    struct RAS_DString *dataPtr;
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
    int fileWidth = 0, fileHeight = 0, fileDepth = 0, fileLength = 0;
    int fileType = 0, fileMapType = 0, fileMapLength = 0;
    unsigned char *pixelPtr, *red = NULL, *green = NULL, *blue = NULL;
    myblock block;

    if (!ReadRASFileHeader(f, &fileWidth, &fileHeight, &fileDepth,
	    &fileLength, &fileType, &fileMapType, &fileMapLength)) {
	Tcl_AppendResult(interp, "couldn't read raw RAS header from file \"",
		fileName, "\"", NULL);
	return TCL_ERROR;
    }
    if ((fileWidth <= 0) || (fileHeight <= 0)) {
	Tcl_AppendResult(interp, "RAS image file \"", fileName,
		"\" has dimension(s) <= 0", (char *) NULL);
	return TCL_ERROR;
    }

    if ((srcX + width) > fileWidth) {
	width = fileWidth - srcX;
    }
    if ((srcY + height) > fileHeight) {
	height = fileHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
	|| (srcX >= fileWidth) || (srcY >= fileHeight)) {
	return TCL_OK;
    }


    Tk_PhotoGetImage(imageHandle, &block.pub);

    if (fileDepth == 32) {
	block.pub.pixelSize = 4;
	block.pub.offset[3] = 3;
    } else {
	block.pub.pixelSize = 3;
	block.pub.offset[3] = 0;
    }
    if (fileType == 3) {
	block.pub.offset[0] = block.pub.pixelSize - 1 - block.pub.offset[0];
	block.pub.offset[1] = block.pub.pixelSize - 1 - block.pub.offset[1];
	block.pub.offset[2] = block.pub.pixelSize - 1 - block.pub.offset[2];
	block.pub.offset[3] = block.pub.pixelSize - 1 - block.pub.offset[3];
    }
    block.pub.width = width;
    block.pub.pitch = block.pub.pixelSize * fileWidth;
    block.pub.height = height;
    block.pub.pixelPtr = (unsigned char *) ckalloc((unsigned) block.pub.pitch *
	    width * block.pub.pixelSize);

/*    if (fileMapType == 1) {
	Fread(buf, 1, 32, f);
    }*/

    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

    ckfree((char *) block.pub.pixelPtr);
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * FileReadRAS --
 *
 *	This procedure is called by the photo image type to read
 *	RAS format data from a file and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	The access position in file f is changed, and new data is
 *	added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

static int
FileReadRAS(interp, f, fileName, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
  return ReadRAS(interp, f, fileName, NULL, formatString, imageHandle,
		 destX, destY, width, height, srcX, srcY);
}


/*
 *----------------------------------------------------------------------
 *
 * StringReadRAS --
 *
 *	This procedure is called by the photo image type to read
 *	RAS format data from a string and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	New data is added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

static int
StringReadRAS(interp, string, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    char *string;
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
  struct RAS_DString data;

  data.data = string;
  data.p = string;
  return ReadRAS(interp, NULL, "", &data, formatString, imageHandle,
		 destX, destY, width, height, srcX, srcY);
}


/*
 *----------------------------------------------------------------------
 *
 * ReadRASFileHeader --
 *
 *	This procedure reads the RAS header from the beginning of a
 *	RAS file and returns information from the header.
 *
 * Results:
 *	The return value is 1 if file "f" appears to start with a valid
 *      RAS header, and 0 otherwise.  If the header is valid,
 *	then *widthPtr and *heightPtr are modified to hold the
 *	dimensions of the image and *numColors holds the number of
 *	colors and byteSize the number of bytes used for 1 pixel.
 *
 * Side effects:
 *	The access position in f advances.
 *
 *----------------------------------------------------------------------
 */

static int
ReadRASFileHeader(f, widthPtr, heightPtr, depthPtr, lengthPtr, typePtr,
	mapTypePtr, mapLengthPtr)
    FILE *f;			/* Image file to read the header from */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here. */
    int *depthPtr, *lengthPtr;
    int *typePtr, *mapTypePtr,  *mapLengthPtr;
{
    unsigned char buf[24];
    int depth, length, type, mapType, mapLength;

    if ((Fread(buf,1,32,f) != 32) || (buf[0] != 89) ||
	    (buf[1] != 166) || (buf[2] != 106) || (buf[3] != 149)) {
	return 0;
    }
    depth = (buf[12]<<24) + (buf[13]<<16) + (buf[14]<<8) + buf[15];
    length = (buf[16]<<24) + (buf[17]<<16) + (buf[18]<<8) + buf[19];
    type = (buf[20]<<24) + (buf[21]<<16) + (buf[22]<<8) + buf[23];
    mapType = (buf[24]<<24) + (buf[25]<<16) + (buf[26]<<8) + buf[27];
    mapLength = (buf[28]<<24) + (buf[29]<<16) + (buf[30]<<8) + buf[31];

    if (((depth != 1) && (depth != 8) && (depth != 24) &&
	    (depth != 32)) || (type>3) || (mapType>2)) {
	return 0;
    }

/*    printf("%d %d %d %d %d\n", depth, length, type, mapType, mapLength);*/

    *widthPtr = (buf[4]<<24) + (buf[5]<<16) + (buf[6]<<8) + buf[7];
    *heightPtr = (buf[8]<<24) + (buf[9]<<16) + (buf[10]<<8) + buf[11];
    *depthPtr = depth;
    *lengthPtr = length;
    *typePtr = type;
    *mapTypePtr = mapType;
    *mapLengthPtr = mapLength;
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * FileWriteRAS
 *
 *	Writes a RAS image to a file. Just calls WriteRAS
 *      with appropriate arguments.
 *
 * Results:
 *	Returns the return value of WriteRAS
 *
 * Side effects:
 *	A file is (hopefully) created on success.
 *
 *----------------------------------------------------------------------
 */
static int
FileWriteRAS(interp, fileName, formatString, blockPtr)
    Tcl_Interp *interp;
    char *fileName;
    char *formatString;
    Tk_PhotoImageBlock *blockPtr;
{
    return WriteRAS(interp, fileName, (Tcl_DString *)NULL, formatString, blockPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * StringWriteRAS
 *
 *	Writes a RAS image to a string. Just calls WriteRAS
 *      with appropriate arguments.
 *
 * Results:
 *	Returns the return value of WriteRAS
 *
 * Side effects:
 *	The Tcl_DString dataPtr is modified on success.
 *
 *----------------------------------------------------------------------
 */
static int	        
StringWriteRAS(interp, dataPtr, formatString, blockPtr) 
    Tcl_Interp *interp;
    Tcl_DString *dataPtr;
    char *formatString;
    Tk_PhotoImageBlock *blockPtr;
{
  return WriteRAS(interp, (char *) NULL, dataPtr, formatString, blockPtr);
}


/*
 * Yes, I know these macros are dangerous. But it should work fine
 */
#define WRITE(buf) { if (f) fputs(buf, f); else Tcl_DStringAppend(dataPtr, buf, -1);}

/*
 *----------------------------------------------------------------------
 *
 * WriteRAS
 *
 *	This procedure writes a RAS image to the file filename 
 *      (if filename != NULL) or to dataPtr.
 *
 * Results:
 *	Returns TCL_OK on success, or TCL_ERROR on error.
 *      Possible failures are:
 *      1. cannot access file (permissions or path not found)
 *      2. TkPhotoGetMask fails to retrieve the region mask
 *         for the image (should not happen)
 *
 * Side effects:
 *	varies (see StringWriteRAS and FileWriteRAS)
 *
 *----------------------------------------------------------------------
 */
static int
WriteRAS(interp, fileName, dataPtr, formatString, blockPtr)
    Tcl_Interp *interp;
    char *fileName;
    Tcl_DString *dataPtr;
    char *formatString;    
    Tk_PhotoImageBlock *blockPtr;
{
    unsigned int x, y, i, found;
    FILE *f = (FILE *) NULL;
    Tcl_HashTable colors;
    Tcl_HashEntry *entry;
    Tcl_HashSearch search;
    unsigned char *pp;
    unsigned int colvalue, ncolors, chars_per_pixel;
    char temp[9], buffer[256];

    /* open the output file (if needed) */
    if (fileName) {
      f = fopen(fileName, "w");
      if (f == (FILE *)NULL) {
	Tcl_AppendResult(interp, ": cannot open file for writing",
		(char *)NULL);
	return TCL_ERROR;
      }
    }


    /* close the file */
    if (f) {
	fclose(f);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * mInit -- --
 *
 *  This procedure initializes a base64 decoder handle
 *
 * Results:
 *  none
 *
 * Side effects:
 *  the base64 handle is initialized
 *
 *----------------------------------------------------------------------
 */

static void
mInit(string, handle)
    char *string;	/* string containing initial mmencoded data */
    MFile *handle;		/* mmdecode "file" handle */
{
    handle->data = string;
    handle->state = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * mWriteInit -- --
 *
 *  This procedure initializes a base64 decoder handle for writing
 *
 * Results:
 *  none
 *
 * Side effects:
 *  the base64 handle is initialized
 *
 *----------------------------------------------------------------------
 */

static void
mWriteInit(buffer, handle)
    Tcl_DString *buffer;
    MFile *handle;		/* mmencode "file" handle */
{
    Tcl_DStringSetLength(buffer, buffer->RAS_SPACEAvl);
    handle->buffer = buffer;
    handle->data = Tcl_DStringValue(buffer);
    handle->state = 0;
    handle->linelength = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Mread --
 *
 *  This procedure is invoked by the RAS file reader as a 
 *  temporary replacement for "fread", to get RAS data out
 *  of a string (using Mgetc).
 *
 * Results:
 *  The return value is the number of characters "read"
 *
 * Side effects:
 *  the base64 handle will change state.
 *
 *----------------------------------------------------------------------
 */

static int
Mread(dst, chunkSize, numChunks, handle)  
   char *dst;		/* where to put the result */
   size_t chunkSize;	/* size of each transfer */
   size_t numChunks;	/* number of chunks */
   MFile *handle;	/* mmdecode "file" handle */
{
   register int i, c;
   int count = chunkSize * numChunks;

   for(i=0; (i<count) && (c=Mgetc(handle)) != RAS_DONE; i++) {
	*dst++ = c;
   }
   return i;
}

/*
 *----------------------------------------------------------------------
 *
 * Mwrite --
 *
 *  This procedure is invoked by the RAS file reader as a 
 *  temporary replacement for "fwrite", to put RAS data
 *  into string (using Mputc).
 *
 * Results:
 *  The return value is the number of characters "written"
 *
 * Side effects:
 *  the base64 handle will change state.
 *
 *----------------------------------------------------------------------
 */

static int
Mwrite(src, chunkSize, numChunks, handle)  
   unsigned char *src;	/* where to get the data */
   size_t chunkSize;	/* size of each transfer */
   size_t numChunks;	/* number of chunks */
   MFile *handle;	/* mmencode "file" handle */
{
    register int i;
    int count = chunkSize * numChunks;
    int curcount = handle->data - Tcl_DStringValue(handle->buffer);
    int bufcount = curcount + count + count/3 +count/52 + 10;

    /* make sure that the DString contains enough RAS_SPACE */
    if (bufcount >= (handle->buffer->RAS_SPACEAvl + 10)) {
	/* We allocate 4k extra, so we don't have to */
	/* reallocate the buffer too often */
	Tcl_DStringSetLength(handle->buffer, bufcount + 4096);
	handle->data = Tcl_DStringValue(handle->buffer) + curcount;
    }

    /* write the data */
    for(i=0; (i<count) && (Mputc(*src++,handle) != RAS_DONE); i++) {
	/* empty loop body */
    }
    return i;
}


/*
 *----------------------------------------------------------------------
 *
 * Mgetc --
 *
 *  This procedure decodes and returns the next byte from a base64
 *  encoded string.
 *
 * Results:
 *  the next byte (or RAS_DONE) is returned.
 *
 * Side effects:
 *  the base64 handle will change state.
 *
 *----------------------------------------------------------------------
 */

static int
Mgetc(handle)
   register MFile *handle;	/* handle containing decoder data and state */
{
    register int c, result = 0;
     
    if (handle->state == RAS_DONE) {
	return(RAS_DONE);
    }

    do {
	c=char64(*handle->data);
	handle->data++;
    } while (c==RAS_SPACE);

    if (c>RAS_SPECIAL) {
	handle->state = RAS_DONE;
	return(handle->state ? handle->c : RAS_DONE);
    }

    switch (handle->state++) {
	case 0:
	   handle->c = c<<2;
	   result = Mgetc(handle);
	   break;
	case 1:
	   result = handle->c | (c>>4);
	   handle->c = (c&0xF)<<4;
	   break;
	case 2:
	   result = handle->c | (c>>2);
	   handle->c = (c&0x3) << 6;
	   break;
	case 3:
	   result = handle->c | c;
	   handle->state = 0;
	   break;
    }
    return(result);
}

/*
 *----------------------------------------------------------------------
 *
 * Mputc --
 *
 *  This procedure encodes and writes the next byte to a base64
 *  encoded string.
 *
 * Results:
 *  the written byte is returned.
 *
 * Side effects:
 *  the base64 handle will change state.
 *
 *----------------------------------------------------------------------
 */

static int
Mputc(c,handle)
   register int c;		/* character to be written */
   register MFile *handle;	/* handle containing decoder data and state */
{
    c &= 0xff;
    switch (handle->state++) {
	case 0:
	    *handle->data++ = base64_table[(c>>2)&63]; break;
	case 1:
	    c |= handle->c<<8;
	    *handle->data++ = base64_table[(c>>4)&63]; break;
	case 2:
	    handle->state = 0;
	    c |= handle->c<<8;
	    *handle->data++ = base64_table[(c>>6)&63];
	    *handle->data++ = base64_table[c&63]; break;
    }
    handle->c = c;
    if (handle->linelength++ > 52) {
	handle->linelength = 0;
	*handle->data++ = '\n';
    }
    return c & 0xff;
}

/*
 *----------------------------------------------------------------------
 *
 * char64 --
 *
 *  This procedure converts a base64 ascii character into
 *  its binary equivalent.  This code is a slightly modified version
 *  of the char64 proc in N. Borenstein's metamail decoder.
 *
 * Results:
 *  the binary value, or an error code.
 *
 * Side effects:
 *  none
 */

static int
char64(c)
int c;
{
    switch(c) {
	case 'A': return(0);  case 'B': return(1);  case 'C': return(2);
	case 'D': return(3);  case 'E': return(4);  case 'F': return(5);
	case 'G': return(6);  case 'H': return(7);  case 'I': return(8);
	case 'J': return(9);  case 'K': return(10); case 'L': return(11);
	case 'M': return(12); case 'N': return(13); case 'O': return(14);
	case 'P': return(15); case 'Q': return(16); case 'R': return(17);
	case 'S': return(18); case 'T': return(19); case 'U': return(20);
	case 'V': return(21); case 'W': return(22); case 'X': return(23);
	case 'Y': return(24); case 'Z': return(25); case 'a': return(26);
	case 'b': return(27); case 'c': return(28); case 'd': return(29);
	case 'e': return(30); case 'f': return(31); case 'g': return(32);
	case 'h': return(33); case 'i': return(34); case 'j': return(35);
	case 'k': return(36); case 'l': return(37); case 'm': return(38);
	case 'n': return(39); case 'o': return(40); case 'p': return(41);
	case 'q': return(42); case 'r': return(43); case 's': return(44);
	case 't': return(45); case 'u': return(46); case 'v': return(47);
	case 'w': return(48); case 'x': return(49); case 'y': return(50);
	case 'z': return(51); case '0': return(52); case '1': return(53);
	case '2': return(54); case '3': return(55); case '4': return(56);
	case '5': return(57); case '6': return(58); case '7': return(59);
	case '8': return(60); case '9': return(61); case '+': return(62);
	case '/': return(63);

	case ' ': case '\t': case '\n': case '\r': case '\f': return(RAS_SPACE);
	case '=':  return(RAS_PAD);
	case '\0': return(RAS_DONE);
	default: return(RAS_BAD);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Fread --
 *
 *  This procedure calls either fread or Mread to read data
 *  from a file or a base64 encoded string.
 *
 * Results: - same as fread
 *
 *----------------------------------------------------------------------
 */

static int
Fread(dst, hunk, count, file)
    unsigned char *dst;		/* where to put the result */
    size_t hunk,count;		/* how many */
    FILE *file;
{
    if (fromData) {
	return(Mread(dst, hunk, count, (MFile *) file));
    } else {
	return(fread((void *)dst, hunk, count, file)); 
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Fwrite --
 *
 *  This procedure calls either fwrite or Mwrite to write data
 *  to a file or to a base64 encoded string.
 *
 * Results: - same as fwrite
 *
 *----------------------------------------------------------------------
 */

static int
Fwrite(src, hunk, count, file)
    unsigned char *src;		/* where to get the data */
    size_t hunk,count;		/* how many */
    FILE *file;
{
    if (fromData) {
	return(Mwrite(src, hunk, count, (MFile *) file));
    } else {
	return(fwrite(src, hunk, count, file)); 
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Fputc --
 *
 *  This procedure calls either fputc or Mputc to write data
 *  to a file or a base64 encoded string.
 *
 * Results: - same as fread
 *
 *----------------------------------------------------------------------
 */

static int
Fputc(c,file)
   register int c;		/* character to be written */
   register FILE *file;		/* handle containing decoder data and state */
{
    if (fromData) {
	return(Mputc(c, (MFile *) file));
    } else {
	return(fputc(c, file)); 
    }
}
