/*
 * imgTIFF.c --
 *
 * A photo image file handler for TIFF files.
 *
 * Uses the libtiff.so library, which is dynamically
 * loaded only when used.
 *
 */

/* Author : Jan Nijtmans */
/* Date   : 3/25/97      */

#include "imgInt.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#ifdef __WIN32__
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define WIN32_LEAN_AND_MEAN
#endif

#ifdef MAC_TCL
#include "libtiff:tiffio.h"
#else
#ifdef HAVE_TIFF_H
#   include <tiffio.h>
#else
#   include "libtiff/tiffio.h"
#endif
#endif

#ifdef __WIN32__
#define TIFF_LIB_NAME "tiff.dll"
#endif

#ifndef TIFF_LIB_NAME
#define TIFF_LIB_NAME "libtiff.so"
#endif

#ifndef TIFF_LIB_SUFFIX
#define TIFF_LIB_SUFFIX "3"
#endif

/*
 * Prototypes for local procedures defined in this file:
 */

static int FileMatchTIFF _ANSI_ARGS_((FILE *f, char *fileName,
	char *formatString, int *widthPtr, int *heightPtr));
static int StringMatchTIFF _ANSI_ARGS_((char *string,
	char *formatString, int *widthPtr, int *heightPtr));
static int FileReadTIFF _ANSI_ARGS_((Tcl_Interp *interp, FILE *f,
	char *fileName, char *formatString, Tk_PhotoHandle imageHandle,
	int destX, int destY, int width, int height, int srcX, int srcY));
static int StringReadTIFF _ANSI_ARGS_((Tcl_Interp *interp, char *string,
	char *formatString, Tk_PhotoHandle imageHandle,
	int destX, int destY, int width, int height, int srcX, int srcY));
static int FileWriteTIFF _ANSI_ARGS_((Tcl_Interp *interp, char *filename,
	char *formatString, Tk_PhotoImageBlock *blockPtr));
static int StringWriteTIFF _ANSI_ARGS_((Tcl_Interp *interp,
	Tcl_DString *dataPtr, char *formatString,
	Tk_PhotoImageBlock *blockPtr));

Tk_PhotoImageFormat imgFmtTIFF = {
	"TIFF",		/* name */
	FileMatchTIFF,	/* fileMatchProc */
	StringMatchTIFF,/* stringMatchProc */
	FileReadTIFF,	/* fileReadProc */
  	StringReadTIFF,	/* stringReadProc */
	FileWriteTIFF,	/* fileWriteProc */
	StringWriteTIFF,/* stringWriteProc */
};

/*
 * We use Tk_ParseArgv to parse any options supplied in the format string.
 */

static char *compression;	/* static variables hold parse results */
				/* ... icky, and not reentrant ... */

static Tk_ArgvInfo writeOptTable[] = {
    {"-compression", TK_ARGV_STRING, "none", (char *) &compression,
	"Select compression method"},
    {NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
	(char *) NULL}
};

static struct TiffFunctions {
    VOID *handle;
    void (* Close) _ANSI_ARGS_((TIFF *));
    int (* GetField) _ANSI_ARGS_((TIFF *, ttag_t, ...));
    int (* GetFieldDefaulted) _ANSI_ARGS_((TIFF *, ttag_t, ...));
    TIFF* (* Open) _ANSI_ARGS_((CONST char*, CONST char*));
    int (* ReadEncodedStrip) _ANSI_ARGS_((TIFF*, tstrip_t, tdata_t, tsize_t));
    int (* ReadRGBAImage) _ANSI_ARGS_((TIFF *, uint32, uint32, uint32*, int));
    int (* ReadTile) _ANSI_ARGS_((TIFF *, uint32, uint32, uint32*, int));
    int (* SetField) _ANSI_ARGS_((TIFF *, ttag_t, ...));
    void (* TileSize) _ANSI_ARGS_((void));
    int (* WriteEncodedStrip) _ANSI_ARGS_((TIFF*, tstrip_t, tdata_t, tsize_t));
    void (* free) _ANSI_ARGS_((tdata_t));
    tdata_t (* malloc) _ANSI_ARGS_((tsize_t));
    TIFFErrorHandler (* SetErrorHandler) _ANSI_ARGS_((TIFFErrorHandler));
    TIFFErrorHandler (* SetWarningHandler) _ANSI_ARGS_((TIFFErrorHandler));
    TIFF* (* ClientOpen) _ANSI_ARGS_((CONST char*, CONST char*, VOID *,
	    TIFFReadWriteProc, TIFFReadWriteProc, TIFFSeekProc,
	    TIFFCloseProc, TIFFSizeProc, TIFFMapFileProc, TIFFUnmapFileProc));
} tiff = {0};

static char *symbols[] = {
    "TIFFClose",
    "TIFFGetField",
    "TIFFGetFieldDefaulted",
    "TIFFOpen",
    "TIFFReadEncodedStrip",
    "TIFFReadRGBAImage",
    "TIFFReadTile",
    "TIFFSetField",
    "TIFFTileSize",
    "TIFFWriteEncodedStrip",
    /* The following symbols are not crucial. If they cannot be
	found, just don't use them. The ClientOpen function is
	more difficult to emulate, but even that is possible. */
    "_TIFFfree",
    "_TIFFmalloc",
    "TIFFSetErrorHandler",
    "TIFFSetWarningHandler",
    "TIFFClientOpen",
    (char *) NULL
};

/*
 * 			 HACK ALERT!!  HACK ALERT!!  HACK ALERT!!
 * This code is hard-wired for reading from files.  In order to read
 * from a data stream, we'll trick fread so we can reuse the same code
 */
 
static int fromData=0;

/*
 * Prototypes for local procedures defined in this file:
 */

static int getint _ANSI_ARGS_((unsigned char *buf, TIFFDataType format,
	int order));
static int CommonReadTIFF _ANSI_ARGS_((Tcl_Interp *interp, TIFF *tif,
	char *formatString, Tk_PhotoHandle imageHandle, int destX, int destY,
	int width, int height, int srcX, int srcY));
static int CommonWriteTIFF _ANSI_ARGS_((Tcl_Interp *interp, TIFF *tif,
	char *formatString, Tk_PhotoImageBlock *blockPtr));
static int load_tiff_library _ANSI_ARGS_((Tcl_Interp *interp));
static void  _TIFFerr    _ANSI_ARGS_((CONST char *, CONST char *, va_list));
static void  _TIFFwarn   _ANSI_ARGS_((CONST char *, CONST char *, va_list));

/*
 * these are for the BASE64 image reader code only
 */

#define TIF_SPECIAL	 (256)
#define TIF_PAD		(TIF_SPECIAL+1)
#define TIF_SPACE	(TIF_SPECIAL+2)
#define TIF_BAD		(TIF_SPECIAL+3)
#define TIF_DONE	(TIF_SPECIAL+4)

static char base64_table[64] = {
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
    'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
    'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
    'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
    'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
    'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
    'w', 'x', 'y', 'z', '0', '1', '2', '3',
    '4', '5', '6', '7', '8', '9', '+', '/'
};

static char *errorMessage = NULL;

typedef struct mFile {
    Tcl_DString *buffer;	/* pointer to dynamical string for writing */
    unsigned char *data;	/* mmencoded source string */
    int c;			/* bits left over from previous character */
    int state;			/* decoder state (0-4 or TIF_DONE) */
    int linelength;		/* length of phisical line already written */
} MFile;

static int		Fread _ANSI_ARGS_((unsigned char *dst, size_t size,
			    size_t count, FILE *file));
static int		Mread _ANSI_ARGS_((unsigned char *dst, size_t size,
			    size_t count, MFile *handle));
static int		Mgetc _ANSI_ARGS_((MFile *handle));
static int		char64 _ANSI_ARGS_((int c));
static void		mInit _ANSI_ARGS_((unsigned char *string,
			    MFile *handle));
static int		Fwrite _ANSI_ARGS_((unsigned char *src, size_t size,
			    size_t count, FILE *file));
static int		Mwrite _ANSI_ARGS_((unsigned char *src, size_t size,
			    size_t count, MFile *handle));
static int		Mputc _ANSI_ARGS_((int c, MFile *handle));
static void		mWriteInit _ANSI_ARGS_((Tcl_DString *buffer,
			    MFile *handle));

static int getint(buf, format, order)
    unsigned char *buf;
    TIFFDataType format;
    int order;
{
    int result;

    switch (format) {
	case TIFF_BYTE:
	    result = buf[0]; break;
	case TIFF_SHORT:
	    result = (buf[order]<<8) + buf[1-order]; break;
	case TIFF_LONG:
	    if (order) {
		result = (buf[3]<<24) + (buf[2]<<16) + (buf[1]<<8) + buf[0];
	    } else {
		result = (buf[0]<<24) + (buf[1]<<16) + (buf[2]<<8) + buf[3];
	    }; break;
	default:
	    result = -1;
    }
    return result;
}

static int
load_tiff_library(interp)
    Tcl_Interp *interp;
{
    if (errorMessage) {
	ckfree(errorMessage);
	errorMessage = NULL;
    }
    if (ImgLoadLib(interp, TIFF_LIB_NAME, TIFF_LIB_SUFFIX, &tiff.handle,
	    symbols, 10) != TCL_OK) {
	return TCL_ERROR;
    }
    if (tiff.SetErrorHandler != NULL) {
	tiff.SetErrorHandler(_TIFFerr);
    }
    if (tiff.SetWarningHandler != NULL) {
	tiff.SetWarningHandler(_TIFFwarn);
    }
    return TCL_OK;
}

static void _TIFFerr(module, fmt, ap)
     const char *module;
     const char *fmt;
     va_list     ap;
{
  char buf[2048];
  char *cp = buf;

  if (module != NULL) {
    sprintf(cp, "%s: ", module);
    cp += strlen(module) + 2;
  }

  vsprintf(cp, fmt, ap);
  if (errorMessage) {
    ckfree(errorMessage);
  }
  errorMessage = (char *) ckalloc(strlen(buf)+1);
  strcpy(errorMessage, buf);
}

/* warnings are not processed in Tcl */
static void _TIFFwarn(module, fmt, ap)
     const char *module;
     const char *fmt;
     va_list     ap;
{
}

/*
 *----------------------------------------------------------------------
 *
 * StringMatchTIFF --
 *
 *  This procedure is invoked by the photo image type to see if
 *  a string contains image data in TIFF format.
 *
 * Results:
 *  The return value is 1 if the first characters in the string
 *  is like TIFF data, and 0 otherwise.
 *
 * Side effects:
 *  the size of the image is placed in widthPre and heightPtr.
 *
 *----------------------------------------------------------------------
 */

static int
StringMatchTIFF(string, formatString, widthPtr, heightPtr)
    char *string;		/* the string containing the image data */
    char *formatString;		/* the image format string */
    int *widthPtr;		/* where to put the string width */
    int *heightPtr;		/* where to put the string height */
{
    int result;
    MFile handle;

    mInit((unsigned char *) string, &handle);
    fromData = 1;
    result = FileMatchTIFF((FILE *) &handle, "inline data",
	formatString, widthPtr, heightPtr);
    fromData = 0;
    return result;
}

static int FileMatchTIFF(f, fileName, formatString, widthPtr, heightPtr)
    FILE *f;
    char *fileName;
    char *formatString;
    int *widthPtr, *heightPtr;
{
    unsigned char buf[4096];
    int i, j, order, w = 0, h = 0;

    i = Fread(buf, 1, 8, f);
    order = (buf[0] == 'I');
    if ((i != 8) || (buf[0] != buf[1])
	    || ((buf[0] != 'I') && (buf[0] != 'M'))
	    || (getint(buf+2,TIFF_SHORT,order) != 42)) {
	return 0;
    }
    i = getint(buf+4,TIFF_LONG,order);

    while (i > 4104) {
	i -= 4096;
	Fread(buf, 1, 4096, f);
    }
    Fread(buf, 1, i-8, f);
    Fread(buf, 1, 2, f);
    i = getint(buf,TIFF_SHORT,order);
    while (i--) {
	Fread(buf, 1, 12, f);
	if (buf[order]!=1) continue;
	j = getint(buf+2,TIFF_SHORT,order);
	j = getint(buf+8, (TIFFDataType) j, order);
	if (buf[1-order]==0) {
	    w = j;
	    if (h>0) break;
	} else if (buf[1-order]==1) {
	    h = j;
	    if (w>0) break;
	}
    }

    if ((w <= 0) || (h <= 0)) {
	return 0;
    }
    *widthPtr = w;
    *heightPtr = h;
    return 1;
}

static int StringReadTIFF(interp, string, formatString, imageHandle,
	destX, destY, width, height, srcX, srcY)
    Tcl_Interp *interp;
    char *string;		/* string containing the image */
    char *formatString;
    Tk_PhotoHandle imageHandle;
    int destX, destY;
    int width, height;
    int srcX, srcY;
{
    TIFF *tif;
    char tempFileName[256];
    int count, result;
    MFile handle;
    char buffer[1024];
    FILE *outfile;

    mInit((unsigned char *) string, &handle);

    tmpnam(tempFileName);
    outfile = fopen(tempFileName,"wb");

    count = Mread(buffer, 1, 1024, &handle);
    while (count == 1024) {
	fwrite(buffer, 1, count, outfile);
	count = Mread(buffer, 1, 1024, &handle);
    }
    if (count>0){
	fwrite(buffer, 1, count, outfile);
    }
    fclose(outfile);

    if (load_tiff_library(interp) != TCL_OK) {
	return TCL_ERROR;
    }

    tif = tiff.Open(tempFileName, "rb");
    if (tif == NULL) {
	unlink(tempFileName);
	return TCL_ERROR;
    }

    result = CommonReadTIFF(interp, tif, formatString, imageHandle,
	    destX, destY, width, height, srcX, srcY);
    unlink(tempFileName);
    return result;
}

static int FileReadTIFF(interp, f, fileName, formatString, imageHandle,
	destX, destY, width, height, srcX, srcY)
    Tcl_Interp *interp;
    FILE *f;
    char *fileName;
    char *formatString;
    Tk_PhotoHandle imageHandle;
    int destX, destY;
    int width, height;
    int srcX, srcY;
{
    TIFF *tif;

    if (load_tiff_library(interp) != TCL_OK) {
	return TCL_ERROR;
    }

    /* The tiff library doesn't use file handles, so
     * there is nothing else to do than close the file
     * and let the library open it again. */

    fclose(f);

    tif = tiff.Open(fileName, "rb");
    if (tif == NULL) {
	printf("error 2\n");
	return TCL_ERROR;
    }

    return CommonReadTIFF(interp, tif, formatString, imageHandle,
	destX, destY, width, height, srcX, srcY);
}

typedef struct myblock {
    Tk_PhotoImageBlock ck;
    int dummy; /* extra space for offset[3], if not included already
		  in Tk_PhotoImageBlock */
} myblock;

#define block bl.ck

static int CommonReadTIFF(interp, tif, formatString, imageHandle,
	destX, destY, width, height, srcX, srcY)
    Tcl_Interp *interp;
    TIFF *tif;
    char *formatString;
    Tk_PhotoHandle imageHandle;
    int destX, destY;
    int width, height;
    int srcX, srcY;
{
    myblock bl;
    unsigned char *pixelPtr = block.pixelPtr;
    uint32 w, h;
    size_t npixels;
    uint32 *raster;

#ifdef WORDS_BIGENDIAN
    block.offset[0] = 3;
    block.offset[1] = 2;
    block.offset[2] = 1;
    block.offset[3] = 3;
#else
    block.offset[0] = 0;
    block.offset[1] = 1;
    block.offset[2] = 2;
    block.offset[3] = 0;
#endif
    block.pixelSize = sizeof (uint32);

    tiff.GetField(tif, TIFFTAG_IMAGEWIDTH, &w);
    tiff.GetField(tif, TIFFTAG_IMAGELENGTH, &h);
    npixels = w * h;
    if (tiff.malloc == NULL) {
	raster = (uint32 *) ckalloc(npixels * sizeof (uint32));
    } else {
	raster = (uint32 *) tiff.malloc(npixels * sizeof (uint32));
    }
    block.width = w;
    block.height = h;
    block.pitch = - (block.pixelSize * (int) w);
    block.pixelPtr = ((unsigned char *) raster) - (h-1) * block.pitch;
    if (raster == NULL) {
	printf("cannot malloc\n");
	return TCL_ERROR;
    }

    if (!tiff.ReadRGBAImage(tif, w, h, raster, 0) || errorMessage) {
	if (tiff.free == NULL) {
	    ckfree((char *)raster);
	} else {
	    tiff.free((char *)raster);
	}
	if (errorMessage) {
	    Tcl_AppendResult(interp, errorMessage, (char *) NULL);
	    ckfree(errorMessage);
	    errorMessage = NULL;
	}
	return TCL_ERROR;
    }

    pixelPtr = block.pixelPtr += srcY * block.pitch
	    + srcX * block.pixelSize;
    ImgPhotoPutBlock(imageHandle, &block, destX,
			destY, width, height);

    if (tiff.free == NULL) {
	ckfree((char *)raster);
    } else {
	tiff.free((char *)raster);
    }
    tiff.Close(tif);
    return TCL_OK;
}

static int StringWriteTIFF(interp, dataPtr, formatString, blockPtr)
    Tcl_Interp *interp;
    Tcl_DString *dataPtr;
    char *formatString;
    Tk_PhotoImageBlock *blockPtr;
{
    TIFF *tif;
    int result;
    MFile handle;
    FILE *f = NULL;
    char buffer[1024];
    char tempFileName[256];

    if (load_tiff_library(interp) != TCL_OK) {
	return TCL_ERROR;
    }

    tmpnam(tempFileName);
    tif = tiff.Open(tempFileName,"w");

    result = CommonWriteTIFF(interp, tif, formatString, blockPtr);
    tiff.Close(tif);
    if (result != TCL_OK) {
	return TCL_ERROR;
    }

    f = fopen(tempFileName,"rb");
    if (f == NULL) {
	Tcl_AppendResult(interp, "error", (char *) NULL);
	return TCL_ERROR;
    }
    mWriteInit(dataPtr,&handle);

    result = fread(buffer, 1, 1024, f);
    while (!feof(f)) {
	Mwrite(buffer, 1, result, &handle);
	result = fread(buffer, 1, 1024, f);
    }
    Mwrite(buffer, 1, result, &handle);
    fclose(f);
    unlink(tempFileName);
    switch(handle.state) {
	case 1:
	    *handle.data++ = base64_table[(handle.c<<4)&63];
	    *handle.data++ = '='; *handle.data++ = '='; break;
	case 2:
	    *handle.data++ = base64_table[(handle.c<<2)&63];
	    *handle.data++ = '='; break;
    }
    Tcl_DStringSetLength(dataPtr,
	    ((char *) handle.data) - Tcl_DStringValue(dataPtr));

    return TCL_OK;
}

static int FileWriteTIFF(interp, filename, formatString, blockPtr)
    Tcl_Interp *interp;
    char *filename;
    char *formatString;
    Tk_PhotoImageBlock *blockPtr;
{
    TIFF *tif;
    int result;
    Tcl_DString nameBuffer; 
    char *fullname;

    if ((fullname=Tcl_TranslateFileName(interp,filename,&nameBuffer))==NULL) {
	return TCL_ERROR;
    }

    if (!(tif = tiff.Open(fullname,"w"))) {
	Tcl_AppendResult(interp, filename, ": ", Tcl_PosixError(interp),
		(char *)NULL);
	Tcl_DStringFree(&nameBuffer);
	return TCL_ERROR;
    }

    Tcl_DStringFree(&nameBuffer);

    if (load_tiff_library(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    result = CommonWriteTIFF(interp, tif, formatString, blockPtr);
    tiff.Close(tif);
    return result;
}

static int CommonWriteTIFF(interp, tif, formatString, blockPtr)
    Tcl_Interp *interp;
    TIFF *tif;
    char *formatString;
    Tk_PhotoImageBlock *blockPtr;
{
    int numsamples, comp;
    unsigned char *data = NULL;

    comp = COMPRESSION_NONE;
    if (formatString != NULL) {
      int argc, length, c;
      char **argv;
      if (Tcl_SplitList(interp, formatString, &argc, &argv) != TCL_OK)
	return TCL_ERROR;
      compression = "none";
      if (Tk_ParseArgv(interp, (Tk_Window) NULL, &argc, argv,
	      writeOptTable, TK_ARGV_NO_LEFTOVERS|TK_ARGV_NO_DEFAULTS)
	      != TCL_OK) {
	ckfree((char *) argv);
	return TCL_ERROR;
      }
      c = compression[0]; length = strlen(compression);
      if ((c == 'n') && (!strncmp(compression,"none",length))) {
	comp = COMPRESSION_NONE;
      } else if ((c == 'r') && (!strncmp(compression,"rle",length))) {
	comp = COMPRESSION_CCITTRLE;
      } else if ((c == 'l') && (!strncmp(compression,"lzw",length))) {
	comp = COMPRESSION_LZW;
      } else if ((c == 'j') && (!strncmp(compression,"jpeg",length))) {
	comp = COMPRESSION_JPEG;
      } else if ((c == 'p') && (!strncmp(compression,"packbits",length))) {
	comp = COMPRESSION_PACKBITS;
      } else if ((c == 'd') && (!strncmp(compression,"deflate",length))) {
	comp = COMPRESSION_DEFLATE;
      } else {
	Tcl_AppendResult(interp, "invalid compression mode \"",
		compression,"\": should be deflate, jpeg, lzw, ",
		"packbits or rle", (char *) NULL);
	ckfree((char *) argv);
	return TCL_ERROR;
      }
      ckfree((char *) argv);
    }
    tiff.SetField(tif, TIFFTAG_IMAGEWIDTH, blockPtr->width);
    tiff.SetField(tif, TIFFTAG_IMAGELENGTH, blockPtr->height);
    tiff.SetField(tif, TIFFTAG_COMPRESSION, comp);

    tiff.SetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    tiff.SetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1);
    tiff.SetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    tiff.SetField(tif, TIFFTAG_ROWSPERSTRIP, blockPtr->height);

    tiff.SetField(tif, TIFFTAG_RESOLUTIONUNIT, (int)2);
    tiff.SetField(tif, TIFFTAG_XRESOLUTION, (float)1200.0);
    tiff.SetField(tif, TIFFTAG_YRESOLUTION, (float)1200.0);

    tiff.SetField(tif, TIFFTAG_BITSPERSAMPLE,   8);
    if ((blockPtr->offset[0] == blockPtr->offset[1])
	    && (blockPtr->offset[0] == blockPtr->offset[2])) {
	numsamples = 1;
	tiff.SetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1);
	tiff.SetField(tif, TIFFTAG_PHOTOMETRIC,    PHOTOMETRIC_MINISBLACK);
    } else {
	numsamples = 3;
	tiff.SetField(tif, TIFFTAG_SAMPLESPERPIXEL, 3);
	tiff.SetField(tif, TIFFTAG_PHOTOMETRIC,     PHOTOMETRIC_RGB);
    }

    if ((blockPtr->pitch == numsamples * blockPtr->width)
	    && (blockPtr->pixelSize == numsamples)) {
	data = blockPtr->pixelPtr;
    } else {
	unsigned char *srcPtr, *dstPtr, *rowPtr;
	int greenOffset, blueOffset, x, y;
	dstPtr = data = (unsigned char *) ckalloc(numsamples *
		blockPtr->width * blockPtr->height);
	rowPtr = blockPtr->pixelPtr + blockPtr->offset[0];
	greenOffset = blockPtr->offset[1] - blockPtr->offset[0];
	blueOffset = blockPtr->offset[2] - blockPtr->offset[0];
	if (blueOffset || greenOffset) {
	    for (y = blockPtr->height; y > 0; y--) {
		srcPtr = rowPtr;
		for (x = blockPtr->width; x>0; x--) {
		    *dstPtr++ = srcPtr[0];
		    *dstPtr++ = srcPtr[greenOffset];
		    *dstPtr++ = srcPtr[blueOffset];
		    srcPtr += blockPtr->pixelSize;
		}
		rowPtr += blockPtr->pitch;
	    }
	} else {
	    for (y = blockPtr->height; y > 0; y--) {
		srcPtr = rowPtr;
		for (x = blockPtr->width; x>0; x--) {
		    *dstPtr++ = srcPtr[0];
		    srcPtr += blockPtr->pixelSize;
		}
		rowPtr += blockPtr->pitch;
	    }
	}
    }

    tiff.WriteEncodedStrip(tif, 0, data,
	    numsamples * blockPtr->width * blockPtr->height);
    if (data != blockPtr->pixelPtr) {
	ckfree((char *) data);
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Minit -- --
 *
 *  This procedure initializes a base64 decoder handle
 *
 * Results:
 *  none
 *
 * Side effects:
 *  the base64 handle is initialized
 *
 *----------------------------------------------------------------------
 */

static void
mInit(string,handle)
   unsigned char *string;	/* string containing initial mmencoded data */
   MFile *handle;		/* mmdecode "file" handle */
{
   handle->data = string;
   handle->state = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * mWriteInit --
 *
 *  This procedure initializes a base64 decoder handle for writing
 *
 * Results:
 *  none
 *
 * Side effects:
 *  the base64 handle is initialized
 *
 *----------------------------------------------------------------------
 */

static void
mWriteInit(buffer, handle)
    Tcl_DString *buffer;
    MFile *handle;		/* mmencode "file" handle */
{
    Tcl_DStringSetLength(buffer, 1024);
    handle->buffer = buffer;
    handle->data = (unsigned char *) Tcl_DStringValue(buffer);
    handle->state = 0;
    handle->linelength = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Mread --
 *
 *  This procedure is invoked by the TIFF file reader as a 
 *  temporary replacement for "fread", to get TIFF data out
 *  of a string (using Mgetc).
 *
 * Results:
 *  The return value is the number of characters "read"
 *
 * Side effects:
 *  the base64 handle will change state.
 *
 *----------------------------------------------------------------------
 */

static int
Mread(dst, chunkSize, numChunks, handle)  
   unsigned char *dst;	/* where to put the result */
   size_t chunkSize;	/* size of each transfer */
   size_t numChunks;	/* number of chunks */
   MFile *handle;	/* mmdecode "file" handle */
{
   register int i, c;
   int count = chunkSize * numChunks;

   for(i=0; i<count && (c=Mgetc(handle)) != TIF_DONE; i++) {
	*dst++ = c;
   }
   return i;
}

/*
 *----------------------------------------------------------------------
 *
 * Mwrite --
 *
 *  This procedure is invoked by the TIFF file reader as a 
 *  temporary replacement for "fwrite", to put TIFF data
 *  into string (using Mputc).
 *
 * Results:
 *  The return value is the number of characters "written"
 *
 * Side effects:
 *  the base64 handle will change state.
 *
 *----------------------------------------------------------------------
 */

static int
Mwrite(src, chunkSize, numChunks, handle)  
    unsigned char *src;	/* where to get the data */
    size_t chunkSize;	/* size of each transfer */
    size_t numChunks;	/* number of chunks */
    MFile *handle;	/* mmencode "file" handle */
{
    register int i;
    int count = chunkSize * numChunks;
    int curcount = ((char *) handle->data) - Tcl_DStringValue(handle->buffer);
    int bufcount = curcount + count + count/3 +count/52 + 1024;

    /* make sure that the DString contains enough space */
    if (bufcount >= (handle->buffer->spaceAvl)) {
	/* We allocate 4k extra, so we don't have to */
	/* reallocate the buffer too often */
	Tcl_DStringSetLength(handle->buffer, bufcount + 4096);
	handle->data = (unsigned char *) Tcl_DStringValue(handle->buffer) + curcount;
    }

    /* write the data */
    for(i=0; (i<count) && (Mputc(*src++,handle) != TIF_DONE); i++) {
	/* empty loop body */
    }
    return i;
}

/*
 *----------------------------------------------------------------------
 *
 * Mgetc --
 *
 *  This procedure decodes and returns the next byte from a base64
 *  encoded string.
 *
 * Results:
 *  The next byte (or TIF_DONE) is returned.
 *
 * Side effects:
 *  The base64 handle will change state.
 *
 *----------------------------------------------------------------------
 */

static int
Mgetc(handle)
   MFile *handle;		/* Handle containing decoder data and state. */
{
    int c;
    int result = 0;		/* Initialization needed only to prevent
				 * gcc compiler warning. */
     
    if (handle->state == TIF_DONE) {
	return(TIF_DONE);
    }

    do {
	c = char64(*handle->data);
	handle->data++;
    } while (c==TIF_SPACE);

    if (c>TIF_SPECIAL) {
	handle->state = TIF_DONE;
	return(handle->state ? handle->c : TIF_DONE);
    }

    switch (handle->state++) {
	case 0:
	   handle->c = c<<2;
	   result = Mgetc(handle);
	   break;
	case 1:
	   result = handle->c | (c>>4);
	   handle->c = (c&0xF)<<4;
	   break;
	case 2:
	   result = handle->c | (c>>2);
	   handle->c = (c&0x3) << 6;
	   break;
	case 3:
	   result = handle->c | c;
	   handle->state = 0;
	   break;
    }
    return(result);
}

/*
 *----------------------------------------------------------------------
 *
 * Mputc --
 *
 *  This procedure encodes and writes the next byte to a base64
 *  encoded string.
 *
 * Results:
 *  the written byte is returned.
 *
 * Side effects:
 *  the base64 handle will change state.
 *
 *----------------------------------------------------------------------
 */

static int
Mputc(c,handle)
    register int c;		/* character to be written */
    register MFile *handle;	/* handle containing decoder data and state */
{
    c &= 0xff;
    switch (handle->state++) {
	case 0:
	    *handle->data++ = base64_table[(c>>2)&63]; break;
	case 1:
	    c |= handle->c<<8;
	    *handle->data++ = base64_table[(c>>4)&63]; break;
	case 2:
	    handle->state = 0;
	    c |= handle->c<<8;
	    *handle->data++ = base64_table[(c>>6)&63];
	    *handle->data++ = base64_table[c&63]; break;
    }
    handle->c = c;
    if (handle->linelength++ > 52) {
	handle->linelength = 0;
	*handle->data++ = '\n';
    }
    return c & 0xff;
}

/*
 *----------------------------------------------------------------------
 *
 * char64 --
 *
 *	This procedure converts a base64 ascii character into its binary
 *	equivalent.  This code is a slightly modified version of the
 *	char64 proc in N. Borenstein's metamail decoder.
 *
 * Results:
 *	The binary value, or an error code.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

static int
char64(c)
int c;
{
    switch(c) {
        case 'A': return(0);  case 'B': return(1);  case 'C': return(2);
        case 'D': return(3);  case 'E': return(4);  case 'F': return(5);
        case 'G': return(6);  case 'H': return(7);  case 'I': return(8);
        case 'J': return(9);  case 'K': return(10); case 'L': return(11);
        case 'M': return(12); case 'N': return(13); case 'O': return(14);
        case 'P': return(15); case 'Q': return(16); case 'R': return(17);
        case 'S': return(18); case 'T': return(19); case 'U': return(20);
        case 'V': return(21); case 'W': return(22); case 'X': return(23);
        case 'Y': return(24); case 'Z': return(25); case 'a': return(26);
        case 'b': return(27); case 'c': return(28); case 'd': return(29);
        case 'e': return(30); case 'f': return(31); case 'g': return(32);
        case 'h': return(33); case 'i': return(34); case 'j': return(35);
        case 'k': return(36); case 'l': return(37); case 'm': return(38);
        case 'n': return(39); case 'o': return(40); case 'p': return(41);
        case 'q': return(42); case 'r': return(43); case 's': return(44);
        case 't': return(45); case 'u': return(46); case 'v': return(47);
        case 'w': return(48); case 'x': return(49); case 'y': return(50);
        case 'z': return(51); case '0': return(52); case '1': return(53);
        case '2': return(54); case '3': return(55); case '4': return(56);
        case '5': return(57); case '6': return(58); case '7': return(59);
        case '8': return(60); case '9': return(61); case '+': return(62);
        case '/': return(63);

	case ' ': case '\t': case '\n': case '\r': case '\f': return(TIF_SPACE);
	case '=':  return(TIF_PAD);
	case '\0': return(TIF_DONE);
	default: return(TIF_BAD);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Fread --
 *
 *  This procedure calls either fread or Mread to read data
 *  from a file or a base64 encoded string.
 *
 * Results: - same as fread
 *
 *----------------------------------------------------------------------
 */

static int
Fread(dst, hunk, count, file)
    unsigned char *dst;		/* where to put the result */
    size_t hunk,count;		/* how many */
    FILE *file;
{
    if (fromData) {
	return(Mread(dst, hunk, count, (MFile *) file));
    } else {
	return(fread((void *)dst, hunk, count, file)); 
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Fwrite --
 *
 *  This procedure calls either fwrite or Mwrite to write data
 *  to a file or to a base64 encoded string.
 *
 * Results: - same as fwrite
 *
 *----------------------------------------------------------------------
 */

static int
Fwrite(src, hunk, count, file)
    unsigned char *src;		/* where to get the data */
    size_t hunk,count;		/* how many */
    FILE *file;
{
    if (fromData) {
	return(Mwrite(src, hunk, count, (MFile *) file));
    } else {
	return(fwrite(src, hunk, count, file)); 
    }
}
